#!/usr/bin/env python

import sys

from fractale.store import FractaleStore
from fractale.subsystem import get_subsystem_solver


def main(args, extra, **kwargs):
    """
    Determine if a jobspec can be satisfied by local resources.
    This is a fairly simple (flat) check.
    """
    store = FractaleStore(args.config_dir)
    solver = get_subsystem_solver(store.clusters_root, args.solver)
    is_satisfied = solver.satisfied(args.jobspec)
    sys.exit(0 if is_satisfied else -1)



#!/usr/bin/env python

import sys

from fractale.store import FractaleStore
from fractale.subsystem import get_subsystem_solver


def main(args, extra, **kwargs):
    """
    Save a cluster and subsystem graph.
    """
    store = FractaleStore(args.config_dir)
    solver = get_subsystem_solver(store.clusters_root, args.solver)
    outfile = args.out
    if not outfile:
        outfile = f"cluster-{args.cluster}-{args.subsystem}-{args.solver}.svg"
    solver.save(args.cluster, args.subsystem, outfile)



from transformers import pipeline

# pip install transformers
# https://pytorch.org/get-started/locally/
# pip3 install torch torchvision torchaudio --index-url https://download.pytorch.org/whl/cu118
pipe = pipeline("text-generation", model="ibm-granite/granite-3b-code-base-128k")

# Load model directly
from transformers import AutoTokenizer, AutoModelForCausalLM
tokenizer = AutoTokenizer.from_pretrained("ibm-granite/granite-3b-code-base-128k")
model = AutoModelForCausalLM.from_pretrained("ibm-granite/granite-3b-code-base-128k")
model.eval()

device = "cpu"
input_text = "def generate():"
# tokenize the text
input_tokens = tokenizer(input_text, return_tensors="pt")
# transfer tokenized inputs to the device
for i in input_tokens:
    input_tokens[i] = input_tokens[i].to(device)
# generate output tokens
output = model.generate(**input_tokens)
# decode output tokens into text
output = tokenizer.batch_decode(output)
# loop over the batch to print, in this example the batch size is 1
for i in output:
    print(i)




import os
from pathlib import Path

import fractale.defaults as defaults
import fractale.utils as utils


class FractaleStore:

    def __init__(self, config_path=None):
        """
        Generate store home.
        """
        self.root = config_path or self.default_root()

    def default_root(self):
        return os.path.join(Path.home(), defaults.fractale_dir)

    @property
    def clusters_root(self):
        return os.path.join(self.root, "clusters")

    def save_subsystem(self, cluster_name, plugin_name, graph):
        """
        Save a graph subsystem
        """
        plugin_path = self.cluster_subsystem(cluster_name, plugin_name)
        if not os.path.exists(plugin_path):
            print(f"Creating subsystem graph store for {plugin_name}")
            os.makedirs(plugin_path)
        print(f"Writing subsystem graph store for {plugin_name} to {plugin_path}")
        utils.write_json(graph, os.path.join(plugin_path, "graph.json"))

    def cluster_subsystem(self, cluster, subsystem):
        return os.path.join(self.clusters_root, cluster.lower(), subsystem)



import os

from .database import DatabaseSolver
from .graph import GraphSolver


def load_solver(backend, path):
    """
    Load the solver backend
    """
    if backend == "database":
        return DatabaseSolver(path)
    if backend == "graph":
        return GraphSolver(path)

    raise ValueError(f"Unsupported backend {backend}")



import re
from io import StringIO

from flux.cli.batch import BatchCmd
from flux.job.directives import DirectiveParser

import fractale.utils as utils
from fractale.transformer.common import JobSpec


class Validator(BatchCmd):
    """
    The validator validates a Flux batch script, and also
    parses arguments into the standard format.
    """

    def derive_failure_reason(self, message):
        """
        Why did the directive parsing fail?
        """
        line = None
        if "line" in message:
            line = int(message.split("line", 1)[-1].split(":", 1)[0])

        # E.g., # # Flux
        if "sentinel changed" in message:
            return "sentinel changed", line

        # Directive after top of script
        if "orphan 'flux:'" in message.lower():
            return "orphan flux", line

        if "unknown directive" in message.lower():
            return "unknown directive", line

        # Always investigate edge cases!
        print("Unseen issue with parsing directive, investigate:")
        print(message)
        import IPython

        IPython.embed()

    def get_directive_parser(self, content, changes=None):
        """
        Read batch script into string, and get directive parser

        If failure is due to a line that can be removed, do it.
        """
        if changes is None:
            changes = []
        string_io = StringIO(content)
        try:
            batchscript = DirectiveParser(string_io)
        except Exception as e:
            string_io.close()
            reason, line = self.derive_failure_reason(" ".join(e.args))
            if line is not None:
                lines = content.split("\n")
                deleted_line = lines[line - 1]
                changes.append({"line": deleted_line, "reason": reason})
                del lines[line - 1]
                return self.get_directive_parser("\n".join(lines), changes)
            else:
                print("The error message did not return a line, take a look why.")
                import IPython

                IPython.embed()
        string_io.close()
        return batchscript, changes

    def unhandled(self, filename):
        return self.parse(filename, return_unhandled=True)

    def parse(self, filename, return_unhandled=False):
        """
        Validate and parse, yielding back arguments.
        """
        content = utils.read_file(filename)
        not_handled = set()

        # Changes are removed lines to get it to read
        batchscript, changes = self.get_directive_parser(content)
        if changes:
            changes = "\n".join(changes)
            raise ValueError(f"Jobspec is invalid, required changes: {changes}")

        # Assume the script is not hashbang, command or directive
        script = [x for x in batchscript.script.split("\n") if not x.startswith("#") and x.strip()]

        # We will populate the common JobSpec
        js = JobSpec(arguments=script)

        # Not parsed yet in flux:
        #   1. input_file (Not sure what this is)
        #   2. I don't think flux has memory per slot
        #   3. I know flux has constraints, add parsed here
        for item in batchscript.directives:
            try:
                # Validation, then mapping to standard
                if item.action == "SETARGS":
                    # This should only be one value, but don't assume
                    for key, value in self.parse_argument_delta(item.args):
                        js = self.update_jobspec(js, key, value, not_handled)

            except Exception:
                name = " ".join(item.args)
                raise ValueError(f"validation failed at {name} line {item.lineno}")
        if return_unhandled:
            return not_handled
        return js

    def update_jobspec(self, js, key, value, unhandled):
        """
        Direct mapping of a key from parsed Flux command line into standard
        """
        if key == "nodes":
            js.num_nodes = value

        # These need additional parsing, and can allow customization
        elif key == "setattr":
            for v in value:
                key, v = v.split("=", 1)
                if key == "container_image":
                    js.container_image = v
                else:
                    js.attrs[key] = v
        elif key == "setopt":
            for v in value:
                key, v = v.split("=", 1)
                js.options[key] = v

        elif key == "cwd":
            js.working_directory = value
        elif key == "nslots":
            js.num_tasks = value
        elif key == "cores_per_task":
            js.cores_per_task = value
        elif key == "gpus_per_task":
            js.gpus_per_slot = value
        elif key == "priority":
            js.priority = value
        elif key == "executable":
            js.executable = value
        elif key == "arguments":
            js.arguments += value
        elif key == "output":
            js.output_file = value
        elif key == "error":
            js.error_file = value
        elif key == "exclusive":
            js.exclusive_access = value
        elif key == "job_name":
            js.job_name = value
        elif key == "env":
            js.env = value
        elif key == "queue":
            js.queue = value
        elif key == "time_limit":
            js.wall_time = parse_time_to_seconds(value)
        elif key == "bank":
            js.account = value
        elif key == "dependency":
            if not js.depends_on:
                js.depends_on = []
            js.depends_on += value
        else:
            print(f"Warning: not handled: {key}={value}")
            unhandled.add(key)
        return js, unhandled

    def parse_argument_delta(self, args):
        """
        Get a single parsed arg by looking at the parser delta.
        """
        defaults = self.parser.parse_args([])
        updated = self.parser.parse_args(args)

        # Find what's different - this will only be one value
        for key, value in vars(updated).items():
            if value != getattr(defaults, key):
                yield key, value


def parse_time_to_seconds(time_str):
    """
    Parses a time string like "1h30m", "1d", "50m", "3600s" into seconds.
    """
    if not time_str:
        return None

    total_seconds = 0
    # Regex to find numbers and their units (d, h, m, s)
    pattern = re.compile(r"(\d+)([dhms])")
    matches = pattern.findall(time_str.lower())

    if not matches and time_str.isdigit():
        return int(time_str)

    for value, unit in matches:
        value = int(value)
        if unit == "d":
            total_seconds += value * 86400
        elif unit == "h":
            total_seconds += value * 3600
        elif unit == "m":
            total_seconds += value * 60
        elif unit == "s":
            total_seconds += value

    return total_seconds if total_seconds > 0 else None


def map_numeric_priority_to_class_name(priority):
    """
    Maps a numerical priority value to a pre-defined Kubernetes PriorityClass name.
    Flux has a default priority of 16. I haven't looked at others yet.
    """
    if priority is None:
        return "normal"

    if priority <= 15:
        return "low"
    elif priority == 16:
        return "normal"
    elif 17 <= priority <= 99:
        return "high"
    else:
        return "urgent"



#!/usr/bin/env python

import os
import sys

import yaml
from rich import print
from rich.pretty import pprint

from fractale.transformer import detect_transformer, get_transformer


def main(args, extra, **kwargs):
    """
    Transform between jobspecs.
    """
    # The jobspec needs to exist as a file here
    if not os.path.exists(args.jobspec):
        sys.exit(f"JobSpec {args.jobspec} does not exist.")

    # If no from transformer defined, try to detect
    if args.from_transformer is None:
        args.from_transformer = detect_transformer(args.jobspec)

    # No selector or solver, just manual transform
    from_transformer = get_transformer(args.from_transformer)
    to_transformer = get_transformer(args.to_transformer)

    # Parse the jobspec to transform from
    normalized_jobspec = from_transformer.parse(args.jobspec)
    final_jobspec = to_transformer.convert(normalized_jobspec)

    if args.pretty and args.to_transformer in ["slurm"]:
        print(final_jobspec)
    elif args.pretty:
        pprint(final_jobspec, indent_guides=True)
    elif args.to_transformer in ["kubernetes"]:
        yaml.dump(final_jobspec, sys.stdout, sort_keys=True, default_flow_style=False)
    else:
        print(final_jobspec)



import fractale.selector.algorithms as algorithms

plugins = {"random": algorithms.random_selection}


def get_selector(name):
    if name not in plugins:
        raise ValueError(f"{name} is not a valid selection algorithm.")
    return plugins[name]



# A cluster has one or more subsystems
create_clusters_sql = """
CREATE TABLE clusters (
  name TEXT PRIMARY KEY NOT NULL
);"""

# A subsystem is owned by a cluster, and has nodes and edges
# We store the type here (defined at the root) for easy query
create_subsystem_sql = """
CREATE TABLE subsystems (
  name TEXT NOT NULL,
  cluster TEXT NOT NULL,
  type TEXT NOT NULL,
  PRIMARY KEY (name, cluster)
);"""

create_nodes_sql = """CREATE TABLE nodes (
  table_id INTEGER PRIMARY KEY,
  subsystem TEXT NOT NULL,
  cluster TEXT NOT NULL,
  label TEXT NOT NULL,
  type TEXT NOT NULL,
  basename TEXT NOT NULL,
  name TEXT NOT NULL,
  id INTEGER NOT NULL,
  FOREIGN KEY(subsystem, cluster) REFERENCES subsystems(name, cluster),
  UNIQUE (cluster, subsystem, label)
);"""

create_attributes_sql = """CREATE TABLE attributes (
  name TEXT NOT NULL,
  subsystem TEXT NOT NULL,
  cluster TEXT NOT NULL,
  value TEXT NOT NULL,
  node TEXT NOT NULL,
  FOREIGN KEY(node) REFERENCES nodes(table_id)
);"""



#!/usr/bin/env python

import os
import sys

from fractale.store import FractaleStore


def main(args, extra, registry):
    """
    Run an extraction. This can be converted to a proper function
    if needed.
    """
    store = FractaleStore(args.config_dir)

    # Discover subsystem plugins
    args.name = args.generate
    plugin = registry.get_plugin(args.generate)

    # Generate / extract based on the plugin type
    # Likely we want to better define these
    attributes = plugin.extract(args, extra)

    # Save to store
    store.save_subsystem(args.cluster, plugin.name, attributes)



#!/usr/bin/env python

import argparse
import os
import sys

from compspec.plugin.parser import plugin_registry
from compspec.plugin.registry import PluginRegistry

import fractale
import fractale.defaults as defaults
from fractale.logger import setup_logger

# Generate the plugin registry to add parsers
registry = PluginRegistry()
registry.discover()


def get_parser():
    parser = argparse.ArgumentParser(
        description="Fractale",
        formatter_class=argparse.RawTextHelpFormatter,
    )

    # Global Variables
    parser.add_argument(
        "--debug",
        dest="debug",
        help="use verbose logging to debug.",
        default=False,
        action="store_true",
    )

    parser.add_argument(
        "--quiet",
        dest="quiet",
        help="suppress additional output.",
        default=False,
        action="store_true",
    )
    parser.add_argument(
        "--config-dir",
        dest="config_dir",
        help="Fractale configuration directory to store subsystems. Defaults to ~/.fractale",
    )
    parser.add_argument(
        "--version",
        dest="version",
        help="show software version.",
        default=False,
        action="store_true",
    )

    subparsers = parser.add_subparsers(
        help="actions",
        title="actions",
        description="actions",
        dest="command",
    )
    subparsers.add_parser("version", description="show software version")

    # This will do the subsystem match before the run
    #    parser.add_argument("name", help="name of subsystem (required)")

    # Generate subsystem metadata for a cluster
    # fractale generate-subsystem --cluster A spack <args>
    generate = subparsers.add_parser(
        "generate",
        formatter_class=argparse.RawTextHelpFormatter,
        description="generate subsystem metadata for a cluster",
    )
    generate.add_argument("-c", "--cluster", help="cluster name")

    # Transform jobspecs from flux to Kubernetes (starting specific)
    transform = subparsers.add_parser(
        "transform",
        formatter_class=argparse.RawTextHelpFormatter,
        description="transform a jobspec",
    )
    transform.add_argument(
        "-t",
        "--to",
        dest="to_transformer",
        help="transform into this jobspec",
        default="kubernetes",
    )
    transform.add_argument(
        "-f",
        "--from",
        dest="from_transformer",
        help="transform from this jobspec",
    )
    transform.add_argument(
        "--pretty",
        help="pretty print in the terminal",
        default=False,
        action="store_true",
    )

    # run.add_argument("-t", "--transform", help="transformer to use", default="flux")
    save = subparsers.add_parser(
        "save",
        formatter_class=argparse.RawTextHelpFormatter,
        description="save a picture of a subsystem graph",
    )
    save.add_argument("cluster", help="cluster to save")
    save.add_argument(
        "--subsystem", help="cluster to save (defaults to containment)", default="containment"
    )
    save.add_argument("--out", help="output file name")

    # This does just the user space subsystem match
    satisfy = subparsers.add_parser(
        "satisfy",
        formatter_class=argparse.RawTextHelpFormatter,
        description="determine clusters that satisfy a jobspec based on user subsystems",
    )
    # Generate a jobspec script
    script = subparsers.add_parser(
        "script",
        formatter_class=argparse.RawTextHelpFormatter,
        description="generate a batch script after satisfy",
    )
    script.add_argument(
        "--selector", help="selection algorithm to use", default="random", choices=["random"]
    )
    script.add_argument(
        "--transformer", help="transformer to use", default="flux", choices=["flux"]
    )
    for cmd in [satisfy, script, transform]:
        cmd.add_argument("jobspec", help="jobspec yaml or json file")

    for cmd in [satisfy, save, script]:
        cmd.add_argument(
            "--solver",
            help="subsystem solved backend",
            default=defaults.solver_backend_default,
            choices=defaults.solver_backends,
        )

    extractors = generate.add_subparsers(
        title="generate",
        description="Use compspec to extract specific application or environment metadata",
        dest="generate",
    )

    # Add plugin parsers to subsystem extractor / generator
    for _, plugin in registry.plugins.items():
        plugin.add_arguments(extractors)
    return parser


def run_fractale():
    """
    this is the main entrypoint.
    """
    parser = get_parser()

    def help(return_code=0):
        version = fractale.__version__

        print("\nFractale v%s" % version)
        parser.print_help()
        sys.exit(return_code)

    # If the user didn't provide any arguments, show the full help
    if len(sys.argv) == 1:
        help()

    # If an error occurs while parsing the arguments, the interpreter will exit with value 2
    args, extra = parser.parse_known_args()

    if args.debug is True:
        os.environ["MESSAGELEVEL"] = "DEBUG"

    # Show the version and exit
    if args.command == "version" or args.version:
        print(fractale.__version__)
        sys.exit(0)

    setup_logger(
        quiet=args.quiet,
        debug=args.debug,
    )

    # Here we can assume instantiated to get args
    if args.command == "generate":
        from .generate_subsystem import main
    elif args.command == "satisfy":
        from .satisfy import main
    elif args.command == "script":
        from .script import main
    elif args.command == "save":
        from .save import main
    elif args.command == "transform":
        from .transform import main
    else:
        help(1)
    global registry
    main(args, extra, registry=registry)


if __name__ == "__main__":
    run_fractale()



import copy
from contextlib import contextmanager

import fractale.utils as utils


def flatten_jobspec_resources(jobspec):
    """
    Given a jobspec, turn the required resources into a flattened version.

    This is intended for subsystem matching that doesn't assume any structure.
    E.g., the database solved backend, which was first just for prototyping.
    """
    resources = {}
    resource_list = copy.deepcopy(jobspec["resources"])
    multiplier = 1

    # Resource lists are nested, under "with"
    while resource_list:
        requires = resource_list.pop(0)
        resource_type = requires["type"]
        resource_count = requires.get("count")

        # The slot is a marker for the set we care about matching
        if resource_type == "slot":
            multiplier = resource_count or 1
        else:
            if resource_type not in resources:
                resources[resource_type] = 0
            resources[resource_type] += resource_count * multiplier
        resource_list += requires.get("with") or []
    return resources


def extract_slot(jobspec):
    """
    Given a jobspec, parse into slot and return counts and requirements.
    This is intended for the containment subsystem.
    """
    js = utils.load_jobspec(jobspec)
    slots = []

    # Tasks will have per slot (this is a list of one item, required)
    # In the future Flux could theoretically support more than one task
    # (and slot) in which case we would need to select the label
    task = js["tasks"][0]
    slot_name = task["slot"]

    # I'm not sure I've ever seen anything other than "per_slot"
    # Might as well raise an error and alert if that happens :)
    slot_count = task.get("count", {})
    if "per_slot" not in slot_count:
        raise ValueError(f"Unexpected value for tasks->slot->count: {slot_count}")

    # Assume a default of one per slot
    slot_count = slot_count.get("per_slot") or 1

    def check_resource(resource, is_slot=False):
        """
        Recursive function to dig into resources under "with.
        Once we find the slot, we save the entire thing (with nesting)
        as a Slot class, which better exposes the counts, etc.
        """
        if "with" in resource:
            for item in resource["with"]:
                if is_slot:
                    new_slot = copy.deepcopy(item)
                    total = item["count"] * slot_count
                    slots.append(Slot(new_slot, total=total))

                # Again, we can eventually support multiple slots by
                # checking the label. Flux right now only expects one.
                check_resource(item, item["type"] == "slot")

    # Kick it off, y'all
    for resource in js["resources"]:
        check_resource(resource, resource["type"] == "slot")
    return slots


class Slot:
    """
    A slot is like a wedge we stick in the containment graph, and say
    "We are defining our needs for the resources below this wedge, and we need
    this many of this wedge." Since we need to evaluate the resource counts
    dynamically, we provide a context-based function to do that. It will restore
    back to the original state when the context is left. The expected usage is:

    # Temporary view of the slot to evaluate with (make changes to)
    with slot.evaluate() as evaluator:
        # This is the current spot we are at in the slot
        requires = evaluator.next_requirement()
        v_type, count = next(requires)

    Check for a StopIteration exception to know when we are done.
    """

    def __init__(self, spec, total=1):
        # This is a nested structure, e.g.,
        # {'type': 'socket', 'count': 1, 'with': [{'type': 'core', 'count': 4}]}
        self.spec = spec
        self.total = total
        # This holds state for what we find during an evaluation
        self._found = {}

    @property
    def start_type(self):
        """
        The start type identifies the top of the slot
        """
        return self.spec["type"]

    @contextmanager
    def evaluate(self):
        """
        Yield a temporary copy of the spec in case it is mutated.
        Also init and reset found, so we can evaluate multiple spots
        for satisfying the same slot instance.
        """
        spec = copy.deepcopy(self.spec)
        self._found = {}
        try:
            yield self
        # Restore the original state, you mangy animal
        finally:
            self.spec = spec
            self._found = {}

    def next_requirement(self):
        """
        Yield requirements. We are strict for now, requiring that we see all levels.
        """
        resources = [copy.deepcopy(self.spec)]
        while resources:
            resource = resources.pop(0)
            yield resource["type"], resource.get("count") or 1
            if "with" in resource:
                resources += resource["with"]

    def count(self, v_type):
        """
        Get the count for a resource type that has been found.
        """
        return self._found.get(v_type) or 0

    def found(self, v_type, count=1, needed=None):
        """
        Indicate that a type was found.
        """
        if v_type not in self._found:
            self._found[v_type] = 0
        self._found[v_type] += 1

        # This tells the caller if we are done with the type.
        if needed is not None and self._found[v_type] == needed:
            return True
        return False

    def satisfied(self):
        """
        Determine if the slot is satisifed, meaning all needed counts
        are <= 0. This should be run in the context of evaluate.
        """
        # These are all the requirements
        needed = {}
        for requires_type, requires_count in self.next_requirement():
            needed[requires_type] = requires_count

        updated = copy.deepcopy(needed)
        for found_type, found_count in self._found.items():
            if found_count >= needed[found_type]:
                del updated[found_type]

        # If updated is empty, we got all needed
        return not updated



import os
import sqlite3

import fractale.defaults as defaults
import fractale.subsystem.queries as queries
import fractale.subsystem.solver as solvers
import fractale.utils as utils
from fractale.logger import LogColors, logger


class SubsystemSolver:
    """
    A subsystem solver has (and loads) one or more subsystems.

    It uses a solver backend to answer questions about resource
    (and subsystem) satisfiability. Different backends (e.g.,
    database or memory graph) are provided.
    """

    def __init__(self, path, backend="database"):
        self.load_solver(backend, path)

    def load_solver(self, backend, path):
        """
        Load a backend that drives the Solver.
        """
        if backend not in defaults.solver_backends:
            raise ValueError(
                f"Backend {backend} is not supported. Options are {defaults.solver_backends}"
            )

        # The subsystem registr
        self.backend = solvers.load_solver(backend, path)

    def render(self, subsystems):
        """
        Render lines for some subsystem result
        """
        return self.backend.render(subsystems)

    def satisfied(self, jobspec, return_results=False):
        """
        Determine if a jobspec is satisfied by user-space subsystems.
        """
        return self.backend.satisfied(jobspec, return_results)

    def save(self, *args, **kwargs):
        """
        Save a graph (or similar graphical output).
        If not implemented, we hit this.
        """
        self.backend.save(*args, **kwargs)


class Subsystem:
    def __init__(self, filename):
        """
        Load a single subsystem
        """
        # Keep track of total counts of things as a quick proxy
        self.counts = {}
        self.load(filename)

    @property
    def metadata(self):
        return self.data["metadata"]

    @property
    def type(self):
        return self.data["metadata"]["type"]

    def iter_nodes(self):
        """
        General function to iterate over nodes depending on if we
        find JGF v1 (list) or JGF v2 (key value pairs).
        """
        if isinstance(self.graph["nodes"], dict):
            for nid, node in self.graph["nodes"].items():
                yield nid, node
        elif isinstance(self.graph["nodes"], list):
            for node in self.graph["nodes"]:
                yield node["id"], node
        else:
            raise ValueError(f"Unsupported subsystem graph type {type(self.graph['nodes'])}")

    def load(self, filename):
        """
        Load a subsystem file, ensuring it exists.
        """
        # Derive the subsystem name from the filepath
        # /home/vanessa/.fractale/clusters/a/spack/graph.json
        # <root>/clusters/<cluster>/<subsystem>/graph.json
        cluster, subsystem = filename.split(os.sep)[-3:-1]
        print(
            f'{LogColors.PURPLE}=> 🍇 Loading cluster "{cluster}" subsystem "{subsystem}"{LogColors.ENDC}'
        )
        self.data = utils.read_json(filename)

        # The name of the subsystem (not the type). E.g., name "spack" has type "software"
        self.name = subsystem
        self.cluster = cluster

        if "graph" not in self.data:
            raise ValueError(f"Subsystem {subsystem} for cluster {cluster} is missing a graph")

        # Nodes are required (edges are not)
        if "nodes" not in self.graph or not self.graph["nodes"]:
            raise ValueError(f"Subsystem {subsystem} for cluster {cluster} is missing nodes")

        # If the metadata doesn't have a type, we assume containment
        if not self.data.get("metadata", {}).get("type"):

            # Flux JGF doesn't have this extra metadata
            if subsystem == "containment":
                metadata = self.data.get("metadata", {})
                metadata["type"] = "containment"
                self.data["metadata"] = metadata
            else:
                raise ValueError(
                    f"Subsystem {subsystem} for cluster {cluster} is missing a type (metadata->type)"
                )

    @property
    def graph(self):
        """
        Return the graph, which is required to exist and be populated to load.
        """
        return self.data["graph"]



import json
import os

import fractale.jobspec as jspec
import fractale.utils as utils
from fractale.logger import LogColors, logger
from fractale.subsystem.match import MatchSet


class Solver:
    """
    A solver determines if one or more subsystems for several
    clusters satisfies a resource requirement
    """

    def save(self, *args, **kwargs):
        """
        Save a graph (or similar graphical output).
        If not implemented, we hit this.
        """
        print(f"Save is not supported for {type(self)}")

    def prepare_requirements(self, jobspec):
        """
        Shared function to read in jobspec, add containment requires
        """
        # This handles json or yaml
        js = utils.load_jobspec(jobspec)

        requires = js["attributes"].get("system", {}).get("requires")

        # Special case: containment - these come from resources in the jobspec
        # we can try matching if we have defined containment subsystems
        requires["containment"] = jspec.flatten_jobspec_resources(js)
        return requires

    def render(self, subsystems):
        """
        Take in a set of cluster matches and
        """
        return []

    def satisfied(self, jobspec, return_results=False):
        """
        Determine if a jobspec is satisfied by user-space subsystems.
        """
        requires = self.prepare_requirements(jobspec)

        # These clusters will satisfy the request
        matches = MatchSet()

        # We don't care about the association with tasks - the requires are matching clusters to entire jobs
        # We could optimize this to be fewer queries, but it's likely trivial for now
        for subsystem_type, items in requires.items():

            # Get one or more matching subsystems (top level) for some number of clusters
            # The subsystem type is like the category (e.g., software)
            subsystems = self.get_subsystem_by_type(subsystem_type)
            if not subsystems:
                continue

            # For each subsystem, since we don't have a query syntax developed, we just look for nodes
            # that have matching attributes. Each here is a tuple, (name, cluster, type)
            for subsystem in subsystems:
                name, cluster, subsystem_type = subsystem

                # If subsystem is containment and we don't have enough totals, fail
                if "containment" in self.subsystems:
                    if not self.assess_containment(requires["containment"]):
                        print(f"{LogColors.RED}=> No Matches due to containment{LogColors.ENDC}")
                        return False

                # "Get nodes in subsystem X" if we have a query syntax we could limit to a type, etc.
                # In this case, the subsystem is the name (e.g., spack) since we might have multiple for
                # a type (e.g., software). This returns labels we can associate with attributes.
                # labels = self.get_subsystem_nodes(cluster, name)

                # "Get attribute key values associated with our search. This is done very stupidly now
                nodes = self.find_nodes(cluster, name, items)
                if not nodes:
                    continue
                # This is adding cluster, subsystem name, match criteria, and node ids
                matches.add(cluster, name, items, nodes)

            if matches:
                print(f"\n{LogColors.OKBLUE}({matches.count}) Matches {LogColors.ENDC}")
                for match in matches.iterset():
                    print(f"cluster ({match.cluster}) subsystem ({match.subsystem})")
                if return_results:
                    return matches
                return True
            else:
                print(f"{LogColors.RED}=> No Matches{LogColors.ENDC}")
            return False

    def load(self, path):
        """
        Load a group of subsystem files, typically json JGF.

        We also are careful to store metadata here that might be needed for
        rendering.
        """
        from fractale.subsystem.subsystem import Subsystem

        self.metadata = {}

        if not os.path.exists(path):
            raise ValueError(f"User subsystem directory {path} does not exist.")
        files = utils.recursive_find(path, "graph[.]json")
        if not files:
            raise ValueError(f"There are no cluster subsystems defined under root {path}")
        for filename in files:
            new_subsystem = Subsystem(filename)
            self.load_subsystem(new_subsystem)
            self.metadata[new_subsystem.name] = new_subsystem.metadata



def find_available_nodes(graph, v_type, v_available, num_nodes_req):
    """
    Finds a specified number of available nodes in the graph.

    This implementation assumes the request is for whole, available nodes.
    It doesn't handle requests for a specific number of cores distributed
    across nodes (which would be more complex).

    Args:
        graph: The graph-tool Graph object.
        v_type: Vertex property map for resource type.
        v_available: Vertex property map for availability.
        num_nodes_req: The number of nodes requested.

    Returns:
        list: A list of available Vertex objects (nodes) satisfying the request,
              or None if the request cannot be satisfied.
    """
    available_nodes = []
    # Iterate through all vertices in the graph
    for v in graph.vertices():
        if v_type[v] == "node" and v_available[v]:
            available_nodes.append(v)

            # If we found enough nodes, return them
            if len(available_nodes) == num_nodes_req:
                return available_nodes  # Found enough

    # If we finished iterating and didn't find enough
    if len(available_nodes) < num_nodes_req:
        print(f"Could not find {num_nodes_req} available nodes. Found only {len(available_nodes)}.")
        return None
    else:
        # This part should technically not be reached due to the check inside loop,
        # but ensures we return the list if the loop finishes exactly on target.
        return available_nodes[:num_nodes_req]


def allocate_nodes(graph, v_available, nodes_to_allocate):
    """
    Marks specified nodes (and their descendants) as unavailable in the graph.

    Args:
        graph: The graph-tool Graph object.
        v_available: Vertex property map for availability (will be modified).
        nodes_to_allocate: A list of Vertex objects (nodes) to mark as unavailable.
    """
    if not nodes_to_allocate:
        print("No nodes provided for allocation.")
        return

    for node_v in nodes_to_allocate:
        # Use BFS/DFS to find all descendants of the node
        # gt.bfs_iterator yields edges, gt.dfs_iterator yields vertices
        for descendant_v in gt.dfs_iterator(graph, node_v):
            # Mark the node itself and all its descendants (sockets, cores) unavailable
            if v_available[descendant_v]:  # Avoid unnecessary writes if already False
                v_available[descendant_v] = False
        print(f"Allocated Node {graph.vp.id[node_v]} and its descendants.")

            # Each valueset is a set of requirements for a node - we assume number is small
            # All requirements (multiple subsystems, for example) must be met for a single cluster.
            for valueset in values:
                # Within a valueset (a dict of properties that must apply to one node)
                # we need to have overlapping vertices. This is stupid and inefficient, but it
                # will get the job done.
                vertices_satisfied = []
                for vertex in self.g.get_vertices():
                    # Assemble the single query for one or more nodes
                    for key, value in valueset.items():   
                        # Node properties are directly on the node, not under metadata or attributes
                        if key in node_properties:
                            v_prop = self.g.vertex_properties.get(key)
                        else:
                            v_prop = self.g.vertex_properties.get(f"attribute.{key}")
                        if not v_prop:
                            print(f"{LogColors.RED}=> No Matches due to unknown property {key}{LogColors.ENDC}")                        
                            return False               
                        # If we get here, we found a property with the key, now we check the value
                        found_value = v_prop[vertex]
                        if not found_value or found_value != value:
                            continue
                            
                    # If we get here, we found a vertex that meets all value criteria
                    vertices_satisfied.append(vertex)


# --- Main Execution ---
if __name__ == "blerg":
    # 1. Create the cluster graph
    g, v_type, v_id, v_available = create_cluster_graph(
        NUM_RACKS, NODES_PER_RACK, SOCKETS_PER_NODE, CORES_PER_SOCKET
    )

    # Optional: Print details of some vertices
    print("\nSample Vertex Details:")
    for i in range(min(15, g.num_vertices())):  # Print first few vertices
        v = g.vertex(i)
        print(f"  Vertex {i}: ID='{v_id[v]}', Type='{v_type[v]}', Available={v_available[v]}")

    # Optional: Visualize the graph (requires cairocffi/pycairo)
    try:
        # Use radial tree layout, good for hierarchies
        # pos = gt.radial_tree_layout(g, g.vertex(0)) # g.vertex(0) is the cluster root
        # Use planar layout if radial fails or looks odd
        pos = gt.arf_layout(g, max_iter=100)

        # Color vertices by type
        color_map = {
            "cluster": "red",
            "rack": "blue",
            "node": "green",
            "socket": "orange",
            "core": "grey",
        }
        v_color = g.new_vertex_property("string")
        for v in g.vertices():
            v_color[v] = color_map.get(v_type[v], "black")  # Default to black if type unknown

        # Draw available nodes brighter, unavailable nodes darker/transparent
        v_fill_color_prop = g.new_vertex_property("vector<double>")
        for v in g.vertices():
            base_color_hex = gt.color_names[color_map.get(v_type[v], "black")]
            if v_available[v]:
                v_fill_color_prop[v] = list(base_color_hex)  # Use original color with full alpha
            else:
                # Make unavailable nodes semi-transparent or darker gray
                # v_fill_color_prop[v] = list(base_color_hex)[:3] + [0.3] # RGBA with low alpha
                v_fill_color_prop[v] = [0.5, 0.5, 0.5, 0.7]  # Dark Gray semi-transparent

        print("\nAttempting to draw graph to 'cluster_graph.png'...")
        gt.graph_draw(
            g,
            pos=pos,
            vertex_text=v_id,  # Label vertices with their ID
            vertex_font_size=8,
            # vertex_fill_color=v_color, # Simple color by type
            vertex_fill_color=v_fill_color_prop,  # Color + Availability
            vertex_size=10,
            edge_pen_width=1,
            output_size=(1200, 1000),
            output="cluster_graph.png",
        )
        print("Graph saved to cluster_graph.png (if drawing libraries are installed).")
    except ImportError:
        print("\nCould not draw graph: cairocffi/pycairo not found or other graphics issue.")
    except Exception as e:
        print(f"\nError during graph drawing: {e}")

    # 2. Define a resource request
    nodes_requested = 5
    print(f"\n--- Resource Request: {nodes_requested} nodes ---")

    # 3. Check if the request can be satisfied
    allocation = find_available_nodes(g, v_type, v_available, nodes_requested)

    # 4. Process the result
    if allocation:
        print(f"\nSuccess! Found {len(allocation)} available nodes:")
        allocated_node_ids = [v_id[node_v] for node_v in allocation]
        print(f"  Allocated Node IDs: {allocated_node_ids}")

        # 5. Simulate Allocation (Modify the graph state)
        print("\nSimulating allocation by marking resources as unavailable...")
        allocate_nodes(g, v_available, allocation)

        # 6. Try another request after allocation
        nodes_requested_2 = 5
        print(f"\n--- Second Resource Request: {nodes_requested_2} nodes ---")
        allocation2 = find_available_nodes(g, v_type, v_available, nodes_requested_2)
        if allocation2:
            print(f"\nSuccess! Found {len(allocation2)} available nodes for the second request:")
            allocated_node_ids_2 = [v_id[node_v] for node_v in allocation2]
            print(f"  Allocated Node IDs: {allocated_node_ids_2}")
        else:
            print(f"\nFailed to satisfy the second request for {nodes_requested_2} nodes.")

        # Optional: Redraw graph after allocation
        try:
            # Recalculate colors based on updated availability
            for v in g.vertices():
                base_color_hex = gt.color_names[color_map.get(v_type[v], "black")]
                if v_available[v]:
                    v_fill_color_prop[v] = list(base_color_hex)
                else:
                    v_fill_color_prop[v] = [0.5, 0.5, 0.5, 0.7]  # Dark Gray semi-transparent

            print(
                "\nAttempting to draw graph after allocation to 'cluster_graph_after_alloc.png'..."
            )
            gt.graph_draw(
                g,
                pos=pos,  # Reuse previous layout position
                vertex_text=v_id,
                vertex_font_size=8,
                vertex_fill_color=v_fill_color_prop,
                vertex_size=10,
                edge_pen_width=1,
                output_size=(1200, 1000),
                output="cluster_graph_after_alloc.png",
            )
            print("Graph saved to cluster_graph_after_alloc.png.")
        except Exception as e:
            print(f"\nError during second graph drawing: {e}")

    else:
        print(f"\nFailed to satisfy the initial request for {nodes_requested} nodes.")

    print("\nScript finished.")



import os
import sqlite3

import fractale.subsystem.queries as queries
from fractale.logger import LogColors, logger

from .base import Solver


class DatabaseSolver(Solver):
    """
    A database solver solves for a cluster based on a simple database.
    """

    def __init__(self, path):
        self.subsystems = {}
        self.conn = sqlite3.connect(":memory:")
        self.create_tables()
        self.load(path)

    def __exit__(self):
        self.close()

    def close(self):
        self.conn.close()

    def create_tables(self):
        """
        Create tables for subsytems, nodes, edges.

        Note that I'm flattening the graph, so edges become attributes for
        nodes so it's easy to query. This is a reasonable first shot over
        implementing an actual graph database.
        """
        cursor = self.conn.cursor()

        # Only save metadata we absolutely need
        # Note I'm not saving edges because we don't use
        # them for anything - we are going to parse them
        # into node attributes instead.
        create_sql = [
            queries.create_subsystem_sql,
            queries.create_clusters_sql,
            queries.create_nodes_sql,
            queries.create_attributes_sql,
        ]
        for sql in create_sql:
            cursor.execute(sql)
        self.conn.commit()

    def load_subsystem(self, subsystem):
        """
        Load a new subsystem to the memory database
        """
        cursor = self.conn.cursor()

        # Create the cluster if it doesn't exist
        values = f"('{subsystem.cluster}')"
        fields = '("name")'
        statement = f"INSERT OR IGNORE INTO clusters {fields} VALUES {values}"
        logger.debug(statement)
        cursor.execute(statement)
        self.conn.commit()

        # Create the subsystem - it should error if already exists
        values = f"('{subsystem.name}', '{subsystem.cluster}', '{subsystem.type}')"
        fields = '("name", "cluster", "type")'
        statement = f"INSERT INTO subsystems {fields} VALUES {values}"
        logger.debug(statement)
        cursor.execute(statement)
        self.conn.commit()

        # These are fields to insert a node and attributes
        node_fields = '("subsystem", "cluster", "label", "type", "basename", "name", "id")'

        # NOTE: we don't create nodes here, e.g., iterate and parse into nodes table
        # This would be subsystem.name, cluster name, nid, type basename, name
        # Instead we just add attributes
        attr_fields = '("cluster", "subsystem", "node", "name", "value")'

        # Keep track of counts of all types
        counts = {}

        # Now all attributes, and also include type because I'm lazy
        for nid, node in subsystem.iter_nodes():
            typ = node["metadata"]["type"]

            # Assume a node is a count of 1
            if typ not in counts:
                counts[typ] = 0
            counts[typ] += 1

            attr_values = f"('{subsystem.cluster}', '{subsystem.name}', '{nid}', 'type', '{typ}')"
            statement = f"INSERT INTO attributes {attr_fields} VALUES {attr_values}"
            cursor.execute(statement)
            for key, value in node["metadata"].get("attributes", {}).items():
                attr_values = (
                    f"('{subsystem.cluster}', '{subsystem.name}', '{nid}', '{key}', '{value}')"
                )
                statement = f"INSERT INTO attributes {attr_fields} VALUES {attr_values}"
                cursor.execute(statement)

        # Note that we aren't doing anything with edges currently.
        self.conn.commit()
        self.subsystems[subsystem.name] = counts

    def get_subsystem_nodes(self, cluster, subsystem):
        """
        Get nodes of a subsystem and cluster

        Technically we could skip labels, but I'm assuming we eventually want
        nodes in this query somewhere.
        """
        statement = (
            f"SELECT label from nodes WHERE subsystem = '{subsystem}' AND cluster = '{cluster}';"
        )
        labels = self.query(statement)
        return [f"'{x[0]}'" for x in labels]

    def render(self, subsystems):
        """
        Yield lines for the transformer.
        """
        for subsystem, items in subsystems.items():
            for item in items:
                # This is actually easier to do than a query!
                if subsystem == "spack":
                    for require in item.requires:
                        item_type = require.get("type")
                        item_name = require.get("name")
                        if item_type == "binary" and item_name is not None:
                            yield f"\nspack load {item_name}"

                        elif subsystem == "environment-modules":
                            item_type = require.get("type")
                            # TODO we need to test if this will with with <.>
                            item_name = require.get("attribute.name")
                            if item_type == "module" and item_name is not None:
                                yield f"\nmodule load {item_name}"

    def find_nodes(self, cluster, name, items):
        """
        Given a list of node labels, find children (attributes)
        that have a specific key/value.
        """
        # Final nodes that satisfy all item requirements
        satisfy = set()

        # Each item is a set of requirements for one NODE. If we cannot satisfy one software
        # requirement the cluster does not match.
        for item in items:
            nodes = set()
            i = 0
            for key, value in item.items():
                statement = f"SELECT * from attributes WHERE cluster = '{cluster}' AND subsystem = '{name}' AND name = '{key}' AND value like '{value}';"
                result = self.query(statement)
                # We don't have any nodes yet, all are contenders
                if i == 0:
                    [nodes.add(x[-1]) for x in result]
                else:
                    new_nodes = {x[-1] for x in result}
                    nodes = nodes.intersection(new_nodes)
                i += 1

                # If we don't have nodes left, the cluster isn't a match
                if not nodes:
                    return

            # If we get down here, we found a matching node for one item requirement
            [satisfy.add(x) for x in nodes]
        return satisfy

    def query(self, statement):
        """
        Issue a query to the database, returning fetchall.
        """
        cursor = self.conn.cursor()
        printed = statement

        # Don't overwhelm the output!
        if len(printed) > 150:
            printed = printed[:150] + "..."
        printed = f"{LogColors.OKCYAN}{printed}{LogColors.ENDC}"
        cursor.execute(statement)
        self.conn.commit()

        # Get results, show query and number of results
        results = cursor.fetchall()
        count = (f"{LogColors.PURPLE}({len(results)}){LogColors.ENDC} ").rjust(20)
        logger.info(count + printed)
        return results

    def assess_containment(self, requires):
        """
        A rough heuirstic to see if the cluster has enough resources
        of specific types.
        """
        for typ, count in requires.items():
            if typ not in self.subsystems.get("containment", {}):
                return False
            have_count = self.subsystems["containment"][typ]
            if have_count < count:
                return False
        return True

    def get_subsystem_by_type(self, subsystem_type, ignore_missing=True):
        """
        Get subsystems based on a type. This will return one or more clusters
        that will be contenders for matching.
        """
        # Check 2: the subsystem exists in our database
        statement = f"SELECT * from subsystems WHERE type = '{subsystem_type}';"
        return self.query(statement)



import os

from .subsystem import SubsystemSolver


def get_subsystem_solver(path, backend="database"):
    """
    Generate a user subsystem registry, where the structure is expected
    to be a set of <cluster>/<subsystem>. For the FractaleStore, this
    is the store.clusters_root.
    """
    if not os.path.exists(path):
        raise ValueError(f"Cluster subsystem root {path} does not exist")

    # Generate the subsystem registry
    return SubsystemSolver(path, backend)



__version__ = "0.0.11"
AUTHOR = "Vanessa Sochat"
AUTHOR_EMAIL = "vsoch@users.noreply.github.com"
NAME = "fractale"
PACKAGE_URL = "https://github.com/compspec/fractale"
KEYWORDS = "cluster, orchestration, transformer, jobspec, flux"
DESCRIPTION = "Jobspec specification and translation layer for cluster work"
LICENSE = "LICENSE"


################################################################################
# Global requirements

# Note that the spack / environment modules plugins are installed automatically.
# This doesn't need to be the case.
INSTALL_REQUIRES = (
    ("jsonschema", {"min_version": None}),
    ("compspec", {"min_version": None}),
    ("compspec-spack", {"min_version": None}),
    ("compspec-modules", {"min_version": None}),
    # Yeah, probably overkill, just being used for printing the scripts
    ("rich", {"min_version": None}),
)

TESTS_REQUIRES = (("pytest", {"min_version": "4.6.2"}),)
INSTALL_REQUIRES_ALL = INSTALL_REQUIRES + TESTS_REQUIRES



import random

# Selection algorithms for choosing clusters
# Each selection algorithm should return a list


def random_selection(choices):
    """
    Select randomly
    """
    return [random.choice(choices)]



import graph_tool.all as gt

import fractale.jobspec as jspec
from fractale.logger import LogColors
from fractale.subsystem.match import MatchSet

from .base import Solver

node_properties = [
    "type",
    "id",
    "basename",
    "label",
    "name",
    "subsystem",
    "cluster",
    "subsystem-type",
]


class GraphSolver(Solver):
    """
    A graph solver solves for a cluster by traversing a graph.
    """

    def __init__(self, path):
        self.subsystems = {}
        self.clusters = set()
        self.vertices = {}
        self.create_graph()
        self.load(path)

    def create_graph(self):
        """
        Create graph for subsytems, nodes, edges.
        """
        self.g = gt.Graph(directed=True)

        # The type will be node, cluster, socket, core, etc.
        # The id will be the identifier, e.g., node0, core0, etc.
        for name in node_properties:
            self.g.vertex_properties[name] = self.g.new_vertex_property("string")

    def load_subsystem(self, subsystem):
        """
        Load a new subsystem to the memory graph (backed by C++ and boost)!
        """
        self.clusters.add(subsystem.cluster)

        # Determine if we've seen the subsystem - we can't add it again
        # In practice this should not happen
        subsystems_known = self.subsystems.get(subsystem.cluster) or {}
        if subsystem.name in subsystems_known:
            raise ValueError(f"Subsystem {subsystem.name} was already added to {subsystem.cluster}")

        # Keep global count of different resource types
        counts = {}

        # Get handles for properties for easiest access
        props = self.g.vertex_properties

        # Keep a lookup of nodes based on the id to create edges
        node_lookup = {}

        # First create all nodes.
        for _, node in subsystem.iter_nodes():

            # Create the node vertex, add properties
            node_v = self.g.add_vertex()
            typ = node["metadata"]["type"]

            # Flux containment has an id at the top level, compspec uses id in metadata
            node_id = node.get("id") or node["metadata"]["id"]

            # This needs to be a lookup across subsystems
            global_id = f"{subsystem.cluster}-{subsystem.name}-{node_id}"

            # Update all properties
            props["cluster"][node_v] = subsystem.cluster
            props["subsystem"][node_v] = subsystem.name
            props["subsystem-type"][node_v] = subsystem.type
            props["type"][node_v] = typ
            props["name"][node_v] = node["metadata"]["name"]
            props["basename"][node_v] = node["metadata"]["basename"]
            props["id"][node_v] = global_id

            # Add attributes, arbitrary key value pair metadata
            # TODO, choose property type based on Python type?
            for key, value in node["metadata"].get("attributes", {}).items():
                if key not in props:
                    props[key] = self.g.new_vertex_property("string")
                props[key][node_v] = value

            # Add the node to the lookup
            node_lookup[node_id] = node_v
            self.vertices[global_id] = node_v

            if typ not in counts:
                counts[typ] = 0
            counts[typ] += 1

        # We only need to add edges for containment subsystem, for now
        # We don't have a use case for others
        if subsystem.name == "containment":
            for edgeset in subsystem.graph.get("edges") or []:
                src = edgeset["source"]
                target = edgeset["target"]
                self.g.add_edge(node_lookup[src], node_lookup[target])

        # Update properties and counts
        for key, value in props.items():
            self.g.vertex_properties[key] = value
        subsystems_known[subsystem.name] = counts
        self.subsystems[subsystem.cluster] = subsystems_known

    def save(self, cluster, subsystem, outpath):
        """
        Custom function to save a view of the graph
        """
        cluster_filter = self.g.new_vertex_property("bool", val=False)
        for v in self.g.vertices():
            if (
                self.g.vertex_properties["cluster"][v] == cluster
                and self.g.vertex_properties["subsystem"][v] == subsystem
            ):
                cluster_filter[v] = True
        view = gt.GraphView(self.g, vfilt=cluster_filter)

        # Assume people do sane things with file extensions
        fmt = outpath.rsplit(".")[-1].lower()
        print(f'{LogColors.OKBLUE}Saving to "{outpath}"{LogColors.ENDC}')
        gt.graph_draw(view, output=outpath, fmt=fmt)

    def satisfied(self, jobspec, return_results=False):
        """
        Determine if a jobspec is satisfied by user-space subsystems.
        """
        requires = self.prepare_requirements(jobspec)

        # First get cluster contenders based on subsystem requirements
        # Note we are passing in a list so we get it back :)
        is_satisfied, matches = self.check_subsystem_satisfies(requires)
        if not is_satisfied:
            return MatchSet()

        # Extract slot resources
        # Flux jobspecs should only have one slot (there is only one task allowed).
        slot = jspec.extract_slot(jobspec)
        if not slot:
            raise ValueError("Did not find slot in jobspec.")
        slot = slot[0]

        # Now we do graph stuff. Let's create a graph view that just includes the clusters
        # This will be greedy - we will stop when we find the first match
        for cluster in matches.clusters:
            if self.check_cluster_satisfies(cluster, slot):
                print(f'{LogColors.OKBLUE}Cluster "{cluster}" is a match{LogColors.ENDC}')
            else:
                matches.remove(cluster)

        if matches.count == 0:
            print(f"{LogColors.RED}=> No Matches{LogColors.ENDC}")
            return False
        if return_results:
            return matches
        return True

    def check_cluster_satisfies(self, cluster, slot):
        """
        Given a cluster name, check if it satifies the containment subsystem,
        which is provided via a slot.
        """
        print(
            f'    {LogColors.PURPLE}=> Exploring cluster "{cluster}" containment subsystem{LogColors.ENDC}'
        )

        # Strategy:
        # 1. Create a view of the graph with just the cluster
        # 2. For that view, also filter down to nodes to explore
        # We can't filter to just nodes, we would lose edges to other types
        cluster_filter = self.g.new_vertex_property("bool", val=False)

        # These will be nodes to start at
        node_vertices = []
        for v in self.g.vertices():
            if self.g.vertex_properties["cluster"][v] == cluster:
                cluster_filter[v] = True

                # Is this vertex additionally a node?
                if self.g.vertex_properties["type"][v] == "node":
                    node_vertices.append(v)

        # This is a view of the entire graph of just nodes
        view = gt.GraphView(self.g, vfilt=cluster_filter)

        # Let's traverse! Ee will be greedy and stop when we get the required slots
        # Also keep count of the number of slots we found
        visited = set()
        slots_found = 0

        # Explore until we fill the required slot
        for vertex in node_vertices:
            if vertex in visited:
                continue
            visited.add(vertex)

            # Keep exploring until we find the slot! We only have one edge type, so
            # just immediately look at the current vertex, then neighbors (and not edges)
            to_explore = [vertex]
            while to_explore:
                # Pop from the end, so we get the last added
                contender = to_explore.pop()

                # This is the type of vertex (e.g., core, node). Does it match the slot?
                v_type = view.vp.type[contender]

                # This means we found the type that satisfies a slot.
                # We need to explore it fully for resources asked for and determine if it's good.
                if v_type == slot.start_type:
                    if self.explore_slot(contender, slot):
                        slots_found += 1

                        # This cluster is a match
                        if slots_found == slot.total:
                            return True

                # Keep exploring looking for slots
                else:
                    to_explore += list(contender.out_neighbors())
        return False

    def explore_slot(self, vertex, slot):
        """
        Evaluate a vertex (and children) for satisfying a slot.
        """
        to_explore = [vertex]

        # Keep track of slot counts we find. The traversal with enforce order

        # This creates a temporary view of the slot we can change
        with slot.evaluate() as evaluator:

            # These are requirements (an interable)
            requires = evaluator.next_requirement()
            requires_type, requires_count = next(requires)

            # Look through vertices. DFS means we go down before we explore more.
            while to_explore:
                vertex = to_explore.pop(0)
                v_type = self.g.vp.type[vertex]

                # This means we are satisfying the slot, keep going through requires
                # We assume one vertex == a count of 1. There is a size attribute but
                # I'm not sure it is used in practice.
                if v_type == requires_type:
                    to_explore += list(vertex.out_neighbors())

                    # Tell the evaluator we found the resource type (not sure I like this name)
                    # If this returns true, we have reached the total we need.
                    if evaluator.found(v_type, 1, requires_count):
                        current_count = evaluator.count(v_type)
                        status = f"   {LogColors.PURPLE}({requires_count}/{current_count}) {LogColors.ENDC}".rjust(
                            25
                        )
                        print(
                            status
                            + f"{LogColors.OKCYAN}satisfied resource {v_type} {LogColors.ENDC}"
                        )

                        # Keep going until we have no more.
                        try:
                            requires_type, requires_count = next(requires)
                        except StopIteration:
                            break
                    else:
                        current_count = evaluator.count(v_type)
                        status = f"   {LogColors.PURPLE}({current_count}/{requires_count}) {LogColors.ENDC}".rjust(
                            25
                        )
                        print(
                            status + f"{LogColors.OKCYAN}found resource {v_type} {LogColors.ENDC}"
                        )

                # For now, we assume this is a failure, because
                # the slot doesn't have the resource type we need
                else:
                    return False

            # Is the evaluator satisfied
            return evaluator.satisfied()

        # We should not get here
        return False

    def render(self, subsystems):
        """
        Yield lines for the transformer.
        """
        for subsystem, items in subsystems.items():
            for item in items:

                # Spack needs a spack load. The "right" way to do this would be
                # to get the initial node back from the solver. But I'm too lazy
                # right now for that, especially if this is just a demo
                for v_id in item.details:
                    v = self.vertices[v_id]
                    item_type = self.g.vp["type"][v]
                    item_name = self.g.vp["name"][v]

                    # Spack template
                    if subsystem == "spack" and item_type == "package" and item_name is not None:
                        yield f"\nspack load {item_name}"

                    # Environment modules template
                    elif (
                        subsystem == "environment-modules"
                        and item_type == "module"
                        and item_name is not None
                    ):
                        yield f"\nmodule load {item_name}"

    def check_subsystem_satisfies(self, requires):
        """
        Determine if subsystems (that are not containment) are satisfied. Unlike a graph
        traversal, since this is based on properties we are doing dumb loop queries.
        Dumb programmer, dumb algorithms, what can I say.
        """
        # Keep track of matching clusters. For this strategy, since we evaluate valuesets for each
        # subsystem, we can only consider a requirement fully satisfied when all valuesets are.
        matches = MatchSet()
        contenders = None

        # Subsystem requirements first
        # Note that these subsystems aren't like spack, modules, etc., they are "software" and more general.
        for subsystem, values in requires.items():
            if subsystem == "containment":
                continue

            # We need to keep track of vertices satisfied (for each valueset)
            satisfied_sets = []

            # Keep track of solutions to send back
            # Each valueset is a set of requirements for a node - we assume number is small
            # All requirements (multiple subsystems, for example) must be met for a single cluster.
            for valueset in values:
                # Within a valueset (a dict of properties that must apply to one node)
                # we need to have overlapping vertices. This is stupid and inefficient, but it
                vertices_satisfied = None

                # Assemble the single query for one or more nodes
                for key, value in valueset.items():
                    # Node properties are directly on the node, not under metadata or attributes
                    if key in node_properties:
                        v_prop = self.g.vertex_properties.get(key)
                    else:
                        v_prop = self.g.vertex_properties.get(f"attribute.{key}")
                    if not v_prop:
                        print(
                            f'{LogColors.RED}=> No Matches due to unknown property "{key}{LogColors.ENDC}"'
                        )
                        return False, matches

                    # Do we have matching vertices anywhere in the graph?
                    vertices = gt.find_vertex(self.g, v_prop, value)
                    if not vertices:
                        print(
                            f'{LogColors.RED}=> No Matches due to missing requirement "{key}={value}"{LogColors.ENDC}'
                        )
                        return False, matches

                    # If we get here, we have vertices. We need to consider vertex and cluster overlap
                    if vertices_satisfied is None:
                        vertices_satisfied = set(vertices)
                    else:
                        # If the current vertices we have don't intersect, we can't satisfy
                        if not vertices_satisfied.intersection(vertices):
                            print(
                                f"{LogColors.RED}=> No Matches due to missing requirement {key}={value}{LogColors.ENDC}"
                            )
                            return False, matches

                        # If we get here, there was an intersection, so add it
                        [
                            vertices_satisfied.add(v)
                            for v in vertices_satisfied.intersection(vertices)
                        ]

                    # Now the same for clusters.
                    # If we haven't defined contenders yet, all here are considered.
                    if contenders is None:
                        contenders = {
                            self.g.vertex_properties["cluster"][v] for v in vertices_satisfied
                        }

                    # If we already have contenders:
                    # 1. Overlap means common nodes (are match vertices are still good)
                    # 2. No overlap - we can't consider cluster because there is a requirement not satisfied
                    else:
                        new_contenders = {
                            self.g.vertex_properties["cluster"][v] for v in vertices_satisfied
                        }
                        if not contenders.intersection(new_contenders):
                            print(
                                f"{LogColors.RED}=> No Matches due to no cluster overlap for {valueset}{LogColors.ENDC}"
                            )
                            return False, matches

                        # If we get here, we still have contender clusters
                        [contenders.add(c) for c in contenders.intersection(new_contenders)]
                        satisfied_sets.append([valueset, vertices_satisfied])

            # If we make it here, we've satisfied all valuesets for a subsystem
            for satisfied_set in satisfied_sets:
                items = []
                for v in satisfied_set[1]:
                    items.append(self.g.vp["id"][v])
                # This is adding cluster, subsystem, valueset, and a list of vertex id
                matches.add(
                    self.g.vp["cluster"][v], self.g.vp["subsystem"][v], satisfied_set[0], items
                )

        # One final check
        if matches.count == 0:
            print(
                f"{LogColors.RED}=> No Matches for any cluster for subsystem requirements{LogColors.ENDC}"
            )
            return False, contenders
        return True, matches



#!/usr/bin/env python

import json
import sys

from rich import print, print_json
from rich.json import JSON
from rich.pretty import pprint

from fractale.store import FractaleStore
from fractale.subsystem import get_subsystem_solver
from fractale.transformer import get_transformer


def main(args, extra, **kwargs):
    """
    Save a cluster and subsystem graph.
    """
    store = FractaleStore(args.config_dir)

    # Prepare selector and transformer
    solver = get_subsystem_solver(store.clusters_root, args.solver)
    # This is probably overloaded, but we need to be able to look up
    # the templating logic for a subsystem
    transformer = get_transformer(args.transformer, args.selector, solver)
    matches = solver.satisfied(args.jobspec, return_results=True)
    if matches.count == 0:
        sys.exit(-1)

    # Select the results to generate. This consolidates matches (which might include different nodes)
    # into clusters and groups of subsystems
    for script in transformer.render(matches, args.jobspec):
        # Generate batch script or jobspec and print
        pprint(script)



import json
from dataclasses import dataclass


@dataclass
class Match:
    """
    A single match is a datum that goes into a match set.
    """

    cluster: str
    subsystem: str
    requires: dict
    details: dict


class MatchSet:
    """
    A MatchSet includes one or more matches of clusters-> subsystems to a spec
    """

    def __init__(self):
        # Lookup by cluster, subsystem, and then requirements and matches.
        # This needs to be enough to return to the called and generate
        # templates to submit jobs.
        self.matches = {}

    @property
    def count(self):
        """
        Return the number of cluster matches.
        """
        return len(self.matches.keys())

    @property
    def clusters(self):
        """
        Return a list of clusters.
        """
        return list(self.matches)

    def all(self):
        """
        Custom function to iterate over matches
        """
        return list(self.iterset())

    def remove(self, cluster):
        """
        Remove a cluster from a match.
        """
        if cluster in self.matches:
            del self.matches[cluster]

    def iterset(self):
        """
        Custom function to iterate over matches
        """
        for cluster, by_subsystem in self.matches.items():
            for subsystem, matches in by_subsystem.items():
                for match in matches:
                    yield match

    def add(self, cluster, subsystem, requires, details):
        """
        Add a match, including cluster, subsystem, requirements, and details.
        """
        new_match = Match(cluster, subsystem, requires, details)
        if cluster not in self.matches:
            self.matches[cluster] = {}
        if subsystem not in self.matches[cluster]:
            self.matches[cluster][subsystem] = []
        # this could be given directly, but I don't want to assume
        # that a cluster and subsystem only has one possible match.
        self.matches[cluster][subsystem].append(new_match)
